from __future__ import annotations as _annotations

from inline_snapshot import snapshot
from pydantic_ai import Agent
from pydantic_ai.messages import ModelRequest, ModelResponse, TextPart, ToolCallPart, ToolReturnPart
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.usage import Usage


def test_init(model_name: str, model: OpenAIModel):
    assert model.model_name == model_name


async def test_stream_text(agent: Agent):
    async with agent.run_stream("Who is the current president of USA?") as result:
        assert not result.is_complete
        assert [c async for c in result.stream_text(debounce_by=0.1)] == snapshot(
            ["The current", "The current president of the USA", "The current president of the USA is Joe Biden."],
        )
        assert result.is_complete
        assert result.usage() == snapshot(
            Usage(requests=1, request_tokens=19, response_tokens=4, total_tokens=23),
        )


async def test_agent(agent: Agent):
    result = await agent.run("How many states are there in USA?")
    messages = result.new_messages()

    assert result.usage() == snapshot(
        Usage(requests=1, request_tokens=34, response_tokens=4, total_tokens=38),
    )

    assert isinstance(messages[-1], ModelResponse)
    assert isinstance(messages[-1].parts[0], TextPart)
    assert messages[-1].parts[0].content == result.data

    assert isinstance(result.data, str)
    assert len(result.data) > 0
    assert result.data == snapshot(
        "50 states.",
    )


async def test_agent_joker(agent_joker: Agent):
    result = await agent_joker.run("Hey! I am Doruk. Tell me a joke.")
    messages = result.new_messages()

    assert result.usage() == snapshot(
        Usage(requests=2, request_tokens=367, response_tokens=42, total_tokens=409),
    )

    # model decides to call tool
    assert isinstance(messages[1], ModelResponse)
    assert isinstance(messages[1].parts[0], ToolCallPart)
    assert messages[1].parts[0].tool_name == "generate_joke"

    # tool responds to model
    assert isinstance(messages[2], ModelRequest)
    assert isinstance(messages[2].parts[0], ToolReturnPart)
    assert messages[2].parts[0].tool_name == "generate_joke"
    assert messages[2].parts[0].content == snapshot("""\
Why don't skeletons fight each other?

They don't have the guts!\
""")

    # model's final response
    assert isinstance(messages[3], ModelResponse)
    assert isinstance(messages[3].parts[0], TextPart)
    assert messages[3].parts[0].content == result.data

    # result
    assert isinstance(result.data, str)
    assert len(result.data) > 0
    assert result.data == snapshot(
        "Sure, here's a joke for you: Why don't skeletons fight each other? They don't have the guts! 😄",
    )



from mlx_lm.utils import GenerationResponse
from openai.types import chat
from pydantic_ai import _utils, usage
from pydantic_ai.messages import ToolCallPart


def map_tool_call(t: ToolCallPart) -> chat.ChatCompletionMessageToolCallParam:
    return chat.ChatCompletionMessageToolCallParam(
        id=_utils.guard_tool_call_id(t=t, model_source="mlx-lm"),
        type="function",
        function={"name": t.tool_name, "arguments": t.args_as_json_str()},
    )


def map_usage(response: GenerationResponse) -> usage.Usage:
    return usage.Usage(
        request_tokens=response.prompt_tokens,
        response_tokens=response.generation_tokens,
        total_tokens=response.prompt_tokens + response.generation_tokens,
        details=None,
    )



from dataclasses import dataclass
from typing import Any, Generic, Iterator, Self, TypeVar

from pydantic_ai import _utils

T = TypeVar("T")


@dataclass
class AsyncStream(Generic[T]):
    """Wraps a synchronous iterator in an asynchronous interface.

    This class allows a synchronous iterator to be treated as an
    asynchronous iterator, enabling iteration in `async for` loops
    and usage within `async with` blocks.

    Example usage:
        async def example():
            sync_iter = iter([1, 2, 3])
            async_stream = AsyncStream(sync_iter)

            async for item in async_stream:
                print(item)

            async with AsyncStream(sync_iter) as stream:
                async for item in stream:
                    print(item)
    """

    _iter: Iterator[T]
    """The underlying synchronous iterator."""

    async def __anext__(self) -> T:
        """Return the next item from the synchronous iterator as if it were asynchronous.

        Calls `_utils.sync_anext` to retrieve the next item from the underlying
        synchronous iterator. If the iterator is exhausted, `StopAsyncIteration`
        is raised.
        """
        return _utils.sync_anext(self._iter)

    def __aiter__(self) -> Self:
        return self

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *_args: Any) -> None:
        pass



from __future__ import annotations as _annotations

import pytest
from pydantic_ai import Agent

from pydantic_ai_mlx_lm import MLXModel


@pytest.fixture
def model() -> MLXModel:
    return MLXModel("mlx-community/Llama-3.2-3B-Instruct-4bit")


@pytest.fixture
def agent(model: MLXModel) -> Agent:
    return Agent(model, system_prompt="You are a chatbot, respond short and concisely.")



from dataclasses import dataclass
from datetime import datetime
from typing import AsyncIterable, AsyncIterator

from mlx_lm.utils import GenerationResponse
from pydantic_ai.messages import ModelResponseStreamEvent
from pydantic_ai.models import StreamedResponse

from .utils import map_usage


@dataclass
class MLXStreamedResponse(StreamedResponse):
    """Implementation of `StreamedResponse` for mlx-lm models."""

    _response: AsyncIterable[GenerationResponse]
    _timestamp: datetime

    async def _get_event_iterator(self) -> AsyncIterator[ModelResponseStreamEvent]:
        async for chunk in self._response:
            self._usage += map_usage(chunk)

            # Handle the text part of the response
            content = chunk.text
            if content:
                yield self._parts_manager.handle_text_delta(vendor_part_id="content", content=content)

            # TODO: Handle tool calls
            # for dtc in choice.delta.tool_calls or []:
            #     maybe_event = self._parts_manager.handle_tool_call_delta(
            #         vendor_part_id=dtc.index,
            #         tool_name=dtc.function and dtc.function.name,
            #         args=dtc.function and dtc.function.arguments,
            #         tool_call_id=dtc.id,
            #     )
            #     if maybe_event is not None:
            #         yield maybe_event

    def timestamp(self) -> datetime:
        return self._timestamp



from __future__ import annotations as _annotations

import pytest
from pydantic_ai import Agent, Tool
from pydantic_ai.models.openai import OpenAIModel

from pydantic_ai_providers import LMStudioProvider, OpenRouterProvider

Provider = LMStudioProvider | OpenRouterProvider


@pytest.fixture
def provider() -> Provider:
    return LMStudioProvider()


@pytest.fixture
def model_name(provider: Provider) -> str:
    if isinstance(provider, LMStudioProvider):
        return "mlx-community/Qwen2.5-7B-Instruct-1M-6bit"
    else:
        return "google/gemini-2.0-flash-001"


@pytest.fixture
def model(model_name: str, provider: Provider) -> OpenAIModel:
    return OpenAIModel(model_name, provider=provider)


@pytest.fixture
def agent(model: OpenAIModel) -> Agent:
    return Agent(model, system_prompt="You are a chatbot, respond short and concisely.")


@pytest.fixture
def agent_joker(model: OpenAIModel, agent: Agent) -> Agent:
    async def generate_joke() -> str:
        """Generate a short joke.

        Returns:
            str: Generated joke.
        """
        result = await agent.run("Tell me a short joke.")
        return result.data

    return Agent(
        model,
        system_prompt="You are a chatbot who loves to tell jokes.",
        tools=[Tool(generate_joke, takes_ctx=False)],
    )



from __future__ import annotations as _annotations

from dataclasses import dataclass
from typing import Literal, Union

from mlx.nn import Module  # pyright: ignore[reportMissingTypeStubs]
from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx_lm.utils import load  # pyright: ignore[reportUnknownVariableType]
from openai.types import chat
from pydantic_ai.models import AgentModel, Model, check_allow_model_requests
from pydantic_ai.tools import ToolDefinition

from .agent_model import MLXAgentModel

KnownMLXModelName = Literal["mlx-community/Llama-3.2-3B-Instruct-4bit"]
"""
For a full list see [link](github.com/ml-explore/mlx-examples/blob/main/llms/README.md#supported-models).
"""

MLXModelName = Union[KnownMLXModelName, str]
"""
Since `mlx-lm` supports lots of models, we explicitly list the most common models but allow any name in the type hints.
"""


@dataclass(init=False)
class MLXModel(Model):
    """A model backend that implements `mlx-lm` for local inference.

    Wrapper around [mlx-lm](https://github.com/ml-explore/mlx-examples/tree/main/llms) to run MLX compatible models locally on Apple Silicon.
    """

    model_name: MLXModelName
    model: Module
    tokenizer: TokenizerWrapper

    def __init__(self, model_name: MLXModelName):
        """Initialize an MLX model.

        Args:
            model_name: The name of the MLX compatible model to use. List of models available at
                github.com/ml-explore/mlx-examples/blob/main/llms/README.md#supported-models
        """
        self.model_name = model_name
        self.model, self.tokenizer = load(model_name)

    async def agent_model(
        self, *, function_tools: list[ToolDefinition], allow_text_result: bool, result_tools: list[ToolDefinition]
    ) -> AgentModel:
        """Create an agent model for function calling."""

        check_allow_model_requests()

        tools = [self._map_tool_definition(r) for r in function_tools]
        if result_tools:
            tools += [self._map_tool_definition(r) for r in result_tools]

        return MLXAgentModel(
            model_name=self.model_name,
            model=self.model,
            tokenizer=self.tokenizer,
            allow_text_result=allow_text_result,
            tools=tools,
        )

    def name(self) -> str:
        return f"mlx-lm:{self.model_name}"

    @staticmethod
    def _map_tool_definition(f: ToolDefinition) -> chat.ChatCompletionToolParam:
        return {
            "type": "function",
            "function": {
                "name": f.name,
                "description": f.description,
                "parameters": f.parameters_json_schema,
            },
        }



from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from itertools import chain
from typing import AsyncIterable, AsyncIterator, Iterable, Literal, overload

from mlx.nn import Module  # pyright: ignore[reportMissingTypeStubs]
from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx_lm.utils import GenerationResponse, generate, stream_generate  # pyright: ignore[reportUnknownVariableType]
from openai.types import chat
from pydantic_ai import _utils
from pydantic_ai._utils import guard_tool_call_id as _guard_tool_call_id
from pydantic_ai.exceptions import UnexpectedModelBehavior
from pydantic_ai.messages import (
    ModelMessage,
    ModelRequest,
    ModelResponse,
    RetryPromptPart,
    SystemPromptPart,
    TextPart,
    ToolCallPart,
    ToolReturnPart,
    UserPromptPart,
)
from pydantic_ai.models import AgentModel, StreamedResponse
from pydantic_ai.settings import ModelSettings
from pydantic_ai.usage import Usage

from .response import MLXStreamedResponse
from .stream import AsyncStream
from .utils import map_tool_call


@dataclass
class MLXAgentModel(AgentModel):
    """Implementation of `AgentModel` for MLX models."""

    model_name: str
    model: Module
    tokenizer: TokenizerWrapper

    allow_text_result: bool
    tools: list[chat.ChatCompletionToolParam]

    async def request(
        self, messages: list[ModelMessage], model_settings: ModelSettings | None
    ) -> tuple[ModelResponse, Usage]:
        """Make a non-streaming request to the model."""

        response = self._completions_create(messages, False, model_settings)
        return self._process_response(response), Usage()

    @asynccontextmanager
    async def request_stream(
        self, messages: list[ModelMessage], model_settings: ModelSettings | None
    ) -> AsyncIterator[StreamedResponse]:
        """Make a streaming request to the model."""

        response = self._completions_create(messages, True, model_settings)
        yield await self._process_streamed_response(response)

    @overload
    def _completions_create(
        self, messages: list[ModelMessage], stream: Literal[False], model_settings: ModelSettings | None
    ) -> str:
        pass

    @overload
    def _completions_create(
        self, messages: list[ModelMessage], stream: Literal[True], model_settings: ModelSettings | None
    ) -> AsyncIterable[GenerationResponse]:
        pass

    def _completions_create(
        self, messages: list[ModelMessage], stream: bool, model_settings: ModelSettings | None
    ) -> str | AsyncIterable[GenerationResponse]:
        """Standalone function to make it easier to override"""

        max_tokens = model_settings.get("max_tokens", 1000) if model_settings else 1000
        conversation = list(chain(*(self._map_message(m) for m in messages)))
        prompt = self.tokenizer.apply_chat_template(  # type: ignore
            conversation=conversation,
            tokenize=False,
            add_generation_prompt=True,
        )

        if not stream:
            return generate(
                model=self.model,
                tokenizer=self.tokenizer,
                prompt=prompt,  # type: ignore
                max_tokens=max_tokens,
            )

        generator = stream_generate(
            model=self.model,
            tokenizer=self.tokenizer,
            prompt=prompt,  # type: ignore
            max_tokens=max_tokens,
        )
        return AsyncStream(generator)

    def _process_response(self, response: str) -> ModelResponse:
        """Process a non-streamed response, and prepare a message to return."""

        return ModelResponse(
            parts=[TextPart(content=response)],
            model_name=self.model_name,
            timestamp=datetime.now(timezone.utc),
        )

    async def _process_streamed_response(self, response: AsyncIterable[GenerationResponse]) -> MLXStreamedResponse:
        """Process a streamed response, and prepare a streaming response to return."""

        peekable_response = _utils.PeekableAsyncStream(response)
        first_chunk = await peekable_response.peek()

        if isinstance(first_chunk, _utils.Unset):
            raise UnexpectedModelBehavior("Streamed response ended without content or tool calls")

        return MLXStreamedResponse(
            _model_name=self.model_name,
            _response=peekable_response,
            _timestamp=datetime.now(timezone.utc),
        )

    def _map_message(self, message: ModelMessage) -> Iterable[chat.ChatCompletionMessageParam]:
        """Just maps a `pydantic_ai.Message` to a `openai.types.ChatCompletionMessageParam`."""

        if isinstance(message, ModelRequest):
            yield from self._map_user_message(message)

        if isinstance(message, ModelResponse):
            texts: list[str] = []
            tool_calls: list[chat.ChatCompletionMessageToolCallParam] = []
            for item in message.parts:
                if isinstance(item, TextPart):
                    texts.append(item.content)

                if isinstance(item, ToolCallPart):
                    tool_calls.append(map_tool_call(item))

            message_param = chat.ChatCompletionAssistantMessageParam(role="assistant")
            if texts:
                # Note: model responses from this model should only have one text item, so the following
                # shouldn't merge multiple texts into one unless you switch models between runs:
                message_param["content"] = "\n\n".join(texts)
            if tool_calls:
                message_param["tool_calls"] = tool_calls
            yield message_param

    def _map_user_message(self, message: ModelRequest) -> Iterable[chat.ChatCompletionMessageParam]:
        for part in message.parts:
            if isinstance(part, SystemPromptPart):
                yield chat.ChatCompletionSystemMessageParam(role="system", content=part.content)

            if isinstance(part, UserPromptPart):
                yield chat.ChatCompletionUserMessageParam(role="user", content=part.content)

            if isinstance(part, ToolReturnPart):
                yield chat.ChatCompletionToolMessageParam(
                    role="tool",
                    tool_call_id=_guard_tool_call_id(t=part, model_source="mlx-lm"),
                    content=part.model_response_str(),
                )

            if isinstance(part, RetryPromptPart):
                if part.tool_name is None:
                    yield chat.ChatCompletionUserMessageParam(role="user", content=part.model_response())
                else:
                    yield chat.ChatCompletionToolMessageParam(
                        role="tool",
                        tool_call_id=_guard_tool_call_id(t=part, model_source="mlx-lm"),
                        content=part.model_response(),
                    )



from __future__ import annotations as _annotations

import os
from typing import overload

from httpx import AsyncClient as AsyncHTTPClient
from openai import AsyncOpenAI
from pydantic_ai.models import cached_async_http_client
from pydantic_ai.providers import Provider

try:
    from openai import AsyncOpenAI
except ImportError as _import_error:  # pragma: no cover
    raise ImportError(
        "Please install `openai` to use the OpenRouter provider, "
        "you can use the `openai` optional group — `pip install 'pydantic-ai-slim[openai]'`"
    ) from _import_error


class OpenRouterProvider(Provider[AsyncOpenAI]):
    """Provider for OpenRouter."""

    @property
    def name(self) -> str:
        return "openrouter"

    @property
    def base_url(self) -> str:
        return "https://openrouter.ai/api/v1"

    @property
    def client(self) -> AsyncOpenAI:
        return self._client

    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, *, api_key: str) -> None: ...

    @overload
    def __init__(self, *, api_key: str, http_client: AsyncHTTPClient) -> None: ...

    @overload
    def __init__(self, *, openai_client: AsyncOpenAI | None = None) -> None: ...

    def __init__(
        self,
        *,
        api_key: str | None = None,
        openai_client: AsyncOpenAI | None = None,
        http_client: AsyncHTTPClient | None = None,
    ) -> None:
        if api_key is None and openai_client is None:
            api_key = os.environ.get("OPENROUTER_API_KEY", "api-key-not-set")

        if openai_client is not None:
            self._client = openai_client
        elif http_client is not None:
            self._client = AsyncOpenAI(base_url=self.base_url, api_key=api_key, http_client=http_client)
        else:
            self._client = AsyncOpenAI(base_url=self.base_url, api_key=api_key, http_client=cached_async_http_client())



from __future__ import annotations as _annotations

from inline_snapshot import snapshot
from mlx.nn import Module  # pyright: ignore[reportMissingTypeStubs]
from mlx_lm.tokenizer_utils import TokenizerWrapper
from pydantic_ai import Agent
from pydantic_ai.messages import ModelResponse, TextPart
from pydantic_ai.usage import Usage

from pydantic_ai_mlx_lm import MLXModel


def test_init(model: MLXModel):
    assert isinstance(model.model, Module)
    assert isinstance(model.tokenizer, TokenizerWrapper)
    assert model.name() == f"mlx-lm:{model.model_name}"


async def test_agent(agent: Agent):
    result = await agent.run("How many states are there in USA?")
    messages = result.new_messages()

    assert result.usage() == snapshot(
        Usage(requests=1, request_tokens=25, response_tokens=385, total_tokens=410),
    )

    assert isinstance(messages[-1], ModelResponse)
    assert isinstance(messages[-1].parts[0], TextPart)
    assert messages[-1].parts[0].content == result.data

    assert isinstance(result.data, str)
    assert len(result.data) > 0
    assert result.data == snapshot(
        "There are 50 states in the United States of America.",
    )


async def test_agent_stream(agent: Agent):
    async with agent.run_stream("Who is the current president of USA?") as result:
        assert not result.is_complete
        assert [c async for c in result.stream_text(debounce_by=0.1)] == snapshot(
            ["As of my knowledge cutoff in 2023, Joe Biden is the President of the United States."],
        )
        assert result.is_complete
        assert result.usage() == snapshot(
            Usage(requests=1, request_tokens=1176, response_tokens=231, total_tokens=1407),
        )
